@@ -1,4 +1,5 @@
-# Copyright (c) 2008-2011, Neotion
+# Copyright (c) 2010-2019 Emmanuel Blot <emmanuel.blot@free.fr>
+# Copyright (c) 2008-2016, Neotion
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -23,55 +24,65 @@
 # NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 # EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-"""Bit field and sequence management
+"""Bit field and sequence management."""
 
-"""
-
-from array import array as Array
-from numbers import Integral
-from pyftdi.misc import is_iterable, xor
-
-__all__ = ['BitSequence', 'BitZSequence', 'BitSequenceError', 'BitField']
-
-# Hints for PyLint:
-#   use map(), use short variable names
-#pylint: disable-msg=W0141
-#pylint: disable-msg=C0103
+from typing import Iterable, List, Optional, Tuple, Union
+from .misc import is_iterable, xor
 
+#pylint: disable-msg=invalid-name
+#pylint: disable-msg=unneeded-not
+#pylint: disable-msg=too-many-branches
+#pylint: disable-msg=too-many-arguments
 
 class BitSequenceError(Exception):
     """Bit sequence error"""
-    pass
 
 
-class BitSequence(object):
-    """Bit sequence manipulation"""
+class BitSequence:
+    """Bit sequence.
 
-    __slots__ = [ '_seq' ]
+       Support most of the common bit operations: or, and, shift, comparison,
+       and conversion from and to integral values.
 
-    def __init__(self, value=None, msb=False, length=0, bytes_=None, msby=True):
-        self._seq = Array('B')
+       Bit sequence objects are iterable.
+
+       Can be initialized with another bit sequence, a integral value,
+       a sequence of bytes or an iterable of common boolean values.
+
+       :param value:  initial value
+       :param msb:    most significant bit first or not
+       :param length: count of signficant bits in the bit sequence
+       :param bytes_: initial value specified as a sequence of bytes
+       :param msby:   most significant byte first or not
+    """
+
+    def __init__(self, value: Union['BitSequence', str, int] = None,
+                 msb: bool = False, length: int = 0,
+                 bytes_: Optional[bytes] = None, msby: bool = True):
+        """Instantiate a new bit sequence.
+        """
+        self._seq = bytearray()
         seq = self._seq
         if value and bytes_:
             raise BitSequenceError("Cannot inialize with both a value and "
                                    "bytes")
         if bytes_:
-            provider = msby and list(bytes_).__iter__() or reversed(bytes_)
+            provider = list(bytes_).__iter__() if msby else reversed(bytes_)
             for byte in provider:
                 if isinstance(byte, str):
                     byte = ord(byte)
                 elif byte > 0xff:
                     raise BitSequenceError("Invalid byte value")
                 b = []
-                for x in xrange(8):
-                    b.append(byte&0x1 and True or False)
+                for _ in range(8):
+                    b.append(bool(byte & 0x1))
                     byte >>= 1
                 if msb:
                     b.reverse()
                 seq.extend(b)
         else:
             value = self._tomutable(value)
-        if isinstance(value, Integral):
+        if isinstance(value, int):
             self._init_from_integer(value, msb, length)
         elif isinstance(value, BitSequence):
             self._init_from_sibling(value, msb)
@@ -83,49 +94,49 @@ class BitSequence(object):
             raise BitSequenceError("Cannot initialize from a %s" % type(value))
         self._update_length(length, msb)
 
-    def sequence(self):
+    def sequence(self) -> bytearray:
         """Return the internal representation as a new mutable sequence"""
-        return Array('B', self._seq)
+        return bytearray(self._seq)
 
-    def reverse(self):
+    def reverse(self) -> 'BitSequence':
         """In-place reverse"""
         self._seq.reverse()
         return self
 
-    def invert(self):
+    def invert(self) -> 'BitSequence':
         """In-place invert sequence values"""
-        self._seq = Array('B', [x^1 for x in self._seq])
+        self._seq = bytearray([x ^ 1 for x in self._seq])
         return self
 
-    def append(self, seq):
+    def append(self, seq) -> 'BitSequence':
         """Concatenate a new BitSequence"""
         if not isinstance(seq, BitSequence):
             seq = BitSequence(seq)
         self._seq.extend(seq.sequence())
         return self
 
-    def lsr(self, count):
-        "Left shift rotate"
+    def lsr(self, count: int) -> None:
+        """Left shift rotate"""
         count %= len(self)
         self._seq[:] = self._seq[count:] + self._seq[:count]
 
-    def rsr(self, count):
-        "Right shift rotate"
+    def rsr(self, count: int) -> None:
+        """Right shift rotate"""
         count %= len(self)
         self._seq[:] = self._seq[-count:] + self._seq[:-count]
 
-    def tobit(self):
+    def tobit(self) -> bool:
         """Degenerate the sequence into a single bit, if possible"""
         if len(self) != 1:
             raise BitSequenceError("BitSequence should be a scalar")
-        return self._seq[0] and True or False
+        return bool(self._seq[0])
 
-    def tobyte(self, msb=False):
+    def tobyte(self, msb: bool = False) -> int:
         """Convert the sequence into a single byte value, if possible"""
         if len(self) > 8:
             raise BitSequenceError("Cannot fit into a single byte")
         byte = 0
-        pos = not msb and -1 or 0
+        pos = -1 if not msb else 0
         # copy the sequence
         seq = self._seq[:]
         while seq:
@@ -133,14 +144,14 @@ class BitSequence(object):
             byte |= seq.pop(pos)
         return byte
 
-    def tobytes(self, msb=False, msby=False):
+    def tobytes(self, msb: bool = False, msby: bool = False) -> bytearray:
         """Convert the sequence into a sequence of byte values"""
         blength = (len(self)+7) & (~0x7)
         sequence = list(self._seq)
         if not msb:
             sequence.reverse()
-        bytes_ = Array('B')
-        for pos in xrange(0, blength, 8):
+        bytes_ = bytearray()
+        for pos in range(0, blength, 8):
             seq = sequence[pos:pos+8]
             byte = 0
             while seq:
@@ -149,10 +160,10 @@ class BitSequence(object):
             bytes_.append(byte)
         if msby:
             bytes_.reverse()
-        return bytes_.tolist()
+        return bytes_
 
     @staticmethod
-    def _tomutable(value):
+    def _tomutable(value: Union[str, Tuple]) -> List:
         """Convert a immutable sequence into a mutable one"""
         if isinstance(value, tuple):
             # convert immutable sequence into a list so it can be popped out
@@ -165,22 +176,22 @@ class BitSequence(object):
                 value = list(value)
         return value
 
-    def _init_from_integer(self, value, msb, length):
+    def _init_from_integer(self, value: int, msb: bool, length: int) -> None:
         """Initialize from any integer value"""
-        l = length or -1
+        bl = length or -1
         seq = self._seq
-        while l:
-            seq.append((value & 1) and True or False)
+        while bl:
+            seq.append(bool(value & 1))
             value >>= 1
             if not value:
                 break
-            l -= 1
+            bl -= 1
         if msb:
             seq.reverse()
 
-    def _init_from_iterable(self, iterable, msb):
+    def _init_from_iterable(self, iterable: Iterable, msb: bool) -> None:
         """Initialize from an iterable"""
-        smap = { '0': 0, '1': 1, False: 0, True: 1, 0: 0, 1: 1 }
+        smap = {'0': 0, '1': 1, False: 0, True: 1, 0: 0, 1: 1}
         seq = self._seq
         try:
             if msb:
@@ -190,7 +201,7 @@ class BitSequence(object):
         except KeyError:
             raise BitSequenceError("Invalid binary character in initializer")
 
-    def _init_from_sibling(self, value, msb):
+    def _init_from_sibling(self, value: 'BitSequence', msb: bool) -> None:
         """Initialize from a fellow object"""
         self._seq = value.sequence()
         if msb:
@@ -200,7 +211,7 @@ class BitSequence(object):
         """If a specific length is specified, extend the sequence as
            expected"""
         if length and (len(self) < length):
-            extra = Array('B', [False] * (length-len(self)))
+            extra = bytearray([False] * (length-len(self)))
             if msb:
                 extra.extend(self._seq)
                 self._seq = extra
@@ -216,31 +227,47 @@ class BitSequence(object):
     def __getitem__(self, index):
         if isinstance(index, slice):
             return self.__class__(value=self._seq[index])
-        else:
-            return self._seq[index]
+        return self._seq[index]
 
     def __setitem__(self, index, value):
-        if not isinstance(value, BitSequence):
-            value = self.__class__(value)
-        else:
+        if isinstance(value, BitSequence):
             if issubclass(value.__class__, self.__class__) and \
                value.__class__ != self.__class__:
                 raise BitSequenceError("Cannot set item with instance of a "
                                        "subclass")
         if isinstance(index, slice):
+            value = self.__class__(value, length=len(self._seq[index]))
             self._seq[index] = value.sequence()
         else:
+            if not isinstance(value, BitSequence):
+                value = self.__class__(value)
             val = value.tobit()
-            if len(self._seq) < index:
-                # auto-resize sequence
-                extra = Array('B', [False] * (index+1-len(self)))
-                self._seq.extend(extra)
+            if index > len(self._seq):
+                raise BitSequenceError("Cannot change the sequence size")
             self._seq[index] = val
 
     def __len__(self):
         return len(self._seq)
 
-    def __cmp__(self, other):
+    def __eq__(self, other):
+        return self._cmp(other) == 0
+
+    def __ne__(self, other):
+        return not self == other
+
+    def __le__(self, other):
+        return not self._cmp(other) <= 0
+
+    def __lt__(self, other):
+        return not self._cmp(other) < 0
+
+    def __ge__(self, other):
+        return not self._cmp(other) >= 0
+
+    def __gt__(self, other):
+        return not self._cmp(other) > 0
+
+    def _cmp(self, other):
         # the bit sequence should be of the same length
         ld = len(self) - len(other)
         if ld:
@@ -258,7 +285,7 @@ class BitSequence(object):
         chunks = []
         srepr = repr(self)
         length = len(self)
-        for i in xrange(0, length, 8):
+        for i in range(0, length, 8):
             if i:
                 j = -i
             else:
@@ -267,9 +294,6 @@ class BitSequence(object):
         return '%d: %s' % (len(self), ' '.join(reversed(chunks)))
 
     def __int__(self):
-        return int(long(self))
-
-    def __long__(self):
         value = 0
         for b in reversed(self._seq):
             value <<= 1
@@ -277,27 +301,27 @@ class BitSequence(object):
         return value
 
     def __and__(self, other):
-        if type(other) is not type(self.__class__()):
+        if not isinstance(other, self.__class__):
             raise BitSequenceError('Need a BitSequence to combine')
         if len(self) != len(other):
             raise BitSequenceError('Sequences must be the same size')
-        return self.__class__(value=map(lambda x, y: x and y,
-                                        self._seq, other.sequence()))
+        return self.__class__(value=list(map(lambda x, y: x and y,
+                                             self._seq, other.sequence())))
 
     def __or__(self, other):
-        if type(other) is not type(self.__class__()):
+        if not isinstance(other, self.__class__):
             raise BitSequenceError('Need a BitSequence to combine')
         if len(self) != len(other):
             raise BitSequenceError('Sequences must be the same size')
-        return self.__class__(value=map(lambda x, y: x or y,
-                                        self._seq, other.sequence()))
+        return self.__class__(value=list(map(lambda x, y: x or y,
+                                             self._seq, other.sequence())))
 
     def __add__(self, other):
-        return self.__class__(value = self._seq + other.sequence())
+        return self.__class__(value=self._seq + other.sequence())
 
     def __ilshift__(self, count):
         count %= len(self)
-        seq = Array('B', [0]*count)
+        seq = bytearray([0]*count)
         seq.extend(self._seq[:-count])
         self._seq = seq
         return self
@@ -309,7 +333,7 @@ class BitSequence(object):
         self._seq = seq
         return self
 
-    def inc(self):
+    def inc(self) -> None:
         """Increment the sequence"""
         for p, b in enumerate(self._seq):
             b ^= True
@@ -317,7 +341,7 @@ class BitSequence(object):
             if b:
                 break
 
-    def dec(self):
+    def dec(self) -> None:
         """Decrement the sequence"""
         for p, b in enumerate(self._seq):
             b ^= True
@@ -325,8 +349,9 @@ class BitSequence(object):
             if not b:
                 break
 
-    def invariant(self):
+    def invariant(self) -> bool:
         """Tells whether all bits of the sequence are of the same value.
+
            Return the value, or ValueError if the bits are not of the same
            value
         """
@@ -343,27 +368,34 @@ class BitSequence(object):
 
 
 class BitZSequence(BitSequence):
-    """Tri-state bit sequence manipulation"""
+    """Tri-state bit sequence manipulation.
 
-    __slots__ = [ '_seq' ]
+       Support most of the BitSequence operations, with an extra high-Z state
 
-    Z = 0xff # maximum byte value
+       :param value:  initial value
+       :param msb:    most significant bit first or not
+       :param length: count of signficant bits in the bit sequence
+    """
+
+    __slots__ = ['_seq']
+
+    Z = 0xff  # maximum byte value
 
     def __init__(self, value=None, msb=False, length=0):
         BitSequence.__init__(self, value=value, msb=msb, length=length)
 
     def invert(self):
-        self._seq = [x in (None, BitZSequence.Z) and BitZSequence.Z or x^1 \
-                        for x in self._seq]
+        self._seq = [x in (None, BitZSequence.Z) and BitZSequence.Z or x ^ 1
+                     for x in self._seq]
         return self
 
     def tobyte(self, msb=False):
-        raise BitSequenceError("Type %s cannot be converted to byte" % \
-                                type(self))
+        raise BitSequenceError("Type %s cannot be converted to byte" %
+                               type(self))
 
     def tobytes(self, msb=False, msby=False):
-        raise BitSequenceError("Type %s cannot be converted to bytes" % \
-                                type(self))
+        raise BitSequenceError("Type %s cannot be converted to bytes" %
+                               type(self))
 
     def matches(self, other):
         if not isinstance(self, BitSequence):
@@ -375,15 +407,15 @@ class BitZSequence(BitSequence):
         for (x, y) in zip(self._seq, other.sequence()):
             if BitZSequence.Z in (x, y):
                 continue
-            if not x is y:
+            if x is not y:
                 return False
         return True
 
     def _init_from_iterable(self, iterable, msb):
         """Initialize from an iterable"""
-        smap = { '0': 0, '1': 1, 'Z': BitZSequence.Z,
-                 False: 0, True: 1, None: BitZSequence.Z,
-                 0: 0, 1: 1, BitZSequence.Z: BitZSequence.Z }
+        smap = {'0': 0, '1': 1, 'Z': BitZSequence.Z,
+                False: 0, True: 1, None: BitZSequence.Z,
+                0: 0, 1: 1, BitZSequence.Z: BitZSequence.Z}
         seq = self._seq
         try:
             if msb:
@@ -393,19 +425,15 @@ class BitZSequence(BitSequence):
         except KeyError:
             raise BitSequenceError("Invalid binary character in initializer")
 
-
     def __repr__(self):
-        smap = { False: '0', True: '1', BitZSequence.Z: 'Z' }
+        smap = {False: '0', True: '1', BitZSequence.Z: 'Z'}
         return ''.join([smap[b] for b in reversed(self._seq)])
 
-    def __long__(self):
+    def __int__(self):
         if BitZSequence.Z in self._seq:
             raise BitSequenceError("High-Z BitSequence cannot be converted to "
                                    "an integral type")
-        return BitSequence.__long__(self)
-
-    def __int__(self):
-        return int(long(self))
+        return BitSequence.__int__(self)
 
     def __cmp__(self, other):
         # the bit sequence should be of the same length
@@ -413,7 +441,7 @@ class BitZSequence(BitSequence):
         if ld:
             return ld
         for n, (x, y) in enumerate(zip(self._seq, other.sequence()), start=1):
-            if not x is y:
+            if x is not y:
                 return n
         return 0
 
@@ -423,13 +451,14 @@ class BitZSequence(BitSequence):
                                    'combine')
         if len(self) != len(other):
             raise BitSequenceError('Sequences must be the same size')
+
         def andz(x, y):
             """Compute the boolean AND operation for a tri-state boolean"""
             if BitZSequence.Z in (x, y):
                 return BitZSequence.Z
-            else:
-                return x and y
-        return self.__class__(value=map(andz, self._seq, other.sequence()))
+            return x and y
+        return self.__class__(
+            value=list(map(andz, self._seq, other.sequence())))
 
     def __or__(self, other):
         if not isinstance(self, BitSequence):
@@ -437,13 +466,14 @@ class BitZSequence(BitSequence):
                                    'combine')
         if len(self) != len(other):
             raise BitSequenceError('Sequences must be the same size')
+
         def orz(x, y):
             """Compute the boolean OR operation for a tri-state boolean"""
             if BitZSequence.Z in (x, y):
                 return BitZSequence.Z
-            else:
-                return x or y
-        return self.__class__(value=map(orz, self._seq, other.sequence()))
+            return x or y
+        return self.__class__(value=list(map(orz, self._seq,
+                                             other.sequence())))
 
     def __rand__(self, other):
         return self.__and__(other)
@@ -455,26 +485,27 @@ class BitZSequence(BitSequence):
         return self.__class__(value=other) + self
 
 
-class BitField(object):
-    """Bitfield manipulation
+class BitField:
+    """Bit field class to access and modify an integral value
+
        Beware the slices does not behave as regular Python slices:
        bitfield[3:5] means b3..b5, NOT b3..b4 as with regular slices
     """
 
-    __slots__ = [ '_val' ]
+    __slots__ = ['_val']
 
     def __init__(self, value=0):
         self._val = value
 
     def to_seq(self, msb=0, lsb=0):
         """Return the BitFiled as a sequence of boolean value"""
-        seq = Array('B')
+        seq = bytearray()
         count = 0
         value = self._val
         while value:
             count += 1
             value >>= 1
-        for x in xrange(lsb, max(msb, count)):
+        for x in range(lsb, max(msb, count)):
             seq.append(bool((self._val >> x) & 1))
         return tuple(reversed(seq))
 
@@ -488,10 +519,9 @@ class BitField(object):
             else:
                 offset = index.start
                 count = index.stop-index.start+1
-            mask = (1<<count)-1
+            mask = (1 << count)-1
             return (self._val >> offset) & mask
-        else:
-            return (self._val >> index) & 1
+        return (self._val >> index) & 1
 
     def __setitem__(self, index, value):
         if isinstance(index, slice):
@@ -503,14 +533,19 @@ class BitField(object):
             else:
                 offset = index.start
                 count = index.stop-index.start+1
-            mask = (1<<count)-1
+            mask = (1 << count)-1
             value = (value & mask) << offset
-            mask = mask << offset
+            mask <<= offset
             self._val = (self._val & ~mask) | value
         else:
-            value = (value&1L)<<index
-            mask = (1L)<<index
+            if isinstance(value, bool):
+                value = int(value)
+            value = (value & int(1)) << index
+            mask = int(1) << index
             self._val = (self._val & ~mask) | value
 
     def __int__(self):
         return self._val
+
+    def __str__(self):
+        return bin(self._val)
