@@ -1,40 +1,50 @@
-# Copyright (c) 2010-2011, Emmanuel Blot <emmanuel.blot@free.fr>
+# Copyright (c) 2010-2020, Emmanuel Blot <emmanuel.blot@free.fr>
+# Copyright (c) 2016, Emmanuel Bouaziz <ebouaziz@free.fr>
 # All rights reserved.
 #
-# This library is free software; you can redistribute it and/or
-# modify it under the terms of the GNU Lesser General Public
-# License as published by the Free Software Foundation; either
-# version 2 of the License, or (at your option) any later version.
+# Redistribution and use in source and binary forms, with or without
+# modification, are permitted provided that the following conditions are met:
+#     * Redistributions of source code must retain the above copyright
+#       notice, this list of conditions and the following disclaimer.
+#     * Redistributions in binary form must reproduce the above copyright
+#       notice, this list of conditions and the following disclaimer in the
+#       documentation and/or other materials provided with the distribution.
+#     * Neither the name of the Neotion nor the names of its contributors may
+#       be used to endorse or promote products derived from this software
+#       without specific prior written permission.
 #
-# This library is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-# Lesser General Public License for more details.
-#
-# You should have received a copy of the GNU Lesser General Public
-# License along with this library; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+# ARE DISCLAIMED. IN NO EVENT SHALL NEOTION BE LIABLE FOR ANY DIRECT, INDIRECT,
+# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
+# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
+# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-import struct
-import time
-from array import array as Array
-from pyftdi.ftdi import Ftdi
-from pyftdi.bits import BitSequence
-from pyftdi.misc import hexline
+"""JTAG support for PyFdti"""
 
-__all__ = ['JtagEngine', 'JtagTool']
+from time import sleep
+from typing import Any, List, Tuple, Union
+from .ftdi import Ftdi
+from .bits import BitSequence
+
+#pylint: disable-msg=invalid-name
 
 
 class JtagError(Exception):
-    """Generic JTAG error"""
+    """Generic JTAG error."""
 
 
-class JtagState(object):
+class JtagState:
     """Test Access Port controller state"""
 
-    def __init__(self, name, modes):
+    def __init__(self, name: str, modes: Tuple[str, str]):
         self.name = name
         self.modes = modes
+        self.exits = [self, self]  # dummy value before initial configuration
 
     def __str__(self):
         return self.name
@@ -42,23 +52,23 @@ class JtagState(object):
     def __repr__(self):
         return self.name
 
-    def setx(self, fstate, tstate):
+    def setx(self, fstate: 'JtagState', tstate: 'JtagState'):
         self.exits = [fstate, tstate]
 
     def getx(self, event):
-        x = event and 1 or 0
+        x = int(bool(event))
         return self.exits[x]
 
-    def is_of(self, mode):
+    def is_of(self, mode: str) -> bool:
         return mode in self.modes
 
 
-class JtagStateMachine(object):
-    """Test Access Port controller state machine"""
+class JtagStateMachine:
+    """Test Access Port controller state machine."""
 
     def __init__(self):
         self.states = {}
-        for s, modes in [('test_logic_reset', ('reset',' idle')),
+        for s, modes in [('test_logic_reset', ('reset', ' idle')),
                          ('run_test_idle', ('idle',)),
                          ('select_dr_scan', ('dr',)),
                          ('capture_dr', ('dr', 'shift', 'capture')),
@@ -80,7 +90,7 @@ class JtagStateMachine(object):
         self['run_test_idle'].setx(self['run_test_idle'],
                                    self['select_dr_scan'])
         self['select_dr_scan'].setx(self['capture_dr'],
-                                   self['select_ir_scan'])
+                                    self['select_ir_scan'])
         self['capture_dr'].setx(self['shift_dr'], self['exit_1_dr'])
         self['shift_dr'].setx(self['shift_dr'], self['exit_1_dr'])
         self['exit_1_dr'].setx(self['pause_dr'], self['update_dr'])
@@ -96,40 +106,43 @@ class JtagStateMachine(object):
         self['pause_ir'].setx(self['pause_ir'], self['exit_2_ir'])
         self['exit_2_ir'].setx(self['shift_ir'], self['update_ir'])
         self['update_ir'].setx(self['run_test_idle'], self['select_dr_scan'])
+        self._current = self['test_logic_reset']
 
-        self.reset()
-
-    def __getitem__(self, name):
+    def __getitem__(self, name: str) -> JtagState:
         return self.states[name]
 
-    def state(self):
+    def state(self) -> JtagState:
         return self._current
 
-    def state_of(self, mode):
+    def state_of(self, mode: str) -> bool:
         return self._current.is_of(mode)
 
     def reset(self):
         self._current = self['test_logic_reset']
 
-    def find_path(self, target, source=None):
+    def find_path(self, target: Union[JtagState, str],
+                  source: Union[JtagState, str, None] = None) \
+                  -> List[JtagState]:
         """Find the shortest event sequence to move from source state to
            target state. If source state is not specified, used the current
            state.
-           Return the list of states, including source and target states"""
+
+           :return: the list of states, including source and target states.
+        """
         if source is None:
             source = self.state()
         if isinstance(source, str):
             source = self[source]
         if isinstance(target, str):
             target = self[target]
-        paths = []
+
         def next_path(state, target, path):
             # this test match the target, path is valid
             if state == target:
                 return path+[state]
             # candidate paths
             paths = []
-            for n,x in enumerate(state.exits):
+            for n, x in enumerate(state.exits):
                 # next state is self (loop around), kill the path
                 if x == state:
                     continue
@@ -142,8 +155,8 @@ class JtagStateMachine(object):
                 if npath:
                     paths.append(npath)
             # keep the shortest path
-            return paths and min([(len(l), l) for l in paths],
-                                  key=lambda x: x[0])[1] or []
+            return min([(len(l), l) for l in paths],
+                       key=lambda x: x[0])[1] if paths else []
         return next_path(source, target, [])
 
     def get_events(self, path):
@@ -151,8 +164,8 @@ class JtagStateMachine(object):
            resulting event sequence allows the JTAG state machine to advance
            from the first state to the last one of the input sequence"""
         events = []
-        for s,d in zip(path[:-1], path[1:]):
-            for e,x in enumerate(s.exits):
+        for s, d in zip(path[:-1], path[1:]):
+            for e, x in enumerate(s.exits):
                 if x == d:
                     events.append(e)
         if len(events) != len(path) - 1:
@@ -164,7 +177,7 @@ class JtagStateMachine(object):
             self._current = self._current.getx(event)
 
 
-class JtagController(object):
+class JtagController:
     """JTAG master of an FTDI device"""
 
     TCK_BIT = 0x01   # FTDI output
@@ -176,7 +189,7 @@ class JtagController(object):
     FTDI_PIPE_LEN = 512
 
     # Private API
-    def __init__(self, trst=False, frequency=3.0E6):
+    def __init__(self, trst: bool = False, frequency: float = 3.0E6):
         """
         trst uses the nTRST optional JTAG line to hard-reset the TAP
           controller
@@ -184,102 +197,98 @@ class JtagController(object):
         self._ftdi = Ftdi()
         self._trst = trst
         self._frequency = frequency
-        self.direction = JtagController.TCK_BIT | \
-                         JtagController.TDI_BIT | \
-                         JtagController.TMS_BIT | \
-                         (self._trst and JtagController.TRST_BIT or 0)
-        self._last = None # Last deferred TDO bit
-        self._write_buff = Array('B')
-
-    def __del__(self):
-        self.close()
+        self.direction = (JtagController.TCK_BIT |
+                          JtagController.TDI_BIT |
+                          JtagController.TMS_BIT |
+                          (self._trst and JtagController.TRST_BIT or 0))
+        self._last = None  # Last deferred TDO bit
+        self._write_buff = bytearray()
 
     # Public API
-    def configure(self, vendor, product, interface):
+    def configure(self, url: str) -> None:
         """Configure the FTDI interface as a JTAG controller"""
-        curfreq = self._ftdi.open_mpsse(vendor, product, interface,
-                                        direction=self.direction,
-                                        #initial=0x0,
-                                        frequency=self._frequency)
+        self._ftdi.open_mpsse_from_url(
+            url, direction=self.direction, frequency=self._frequency)
         # FTDI requires to initialize all GPIOs before MPSSE kicks in
-        cmd = Array('B', [Ftdi.SET_BITS_LOW, 0x0, self.direction])
+        cmd = bytearray((Ftdi.SET_BITS_LOW, 0x0, self.direction))
         self._ftdi.write_data(cmd)
 
-    def close(self):
-        if self._ftdi:
+    def close(self) -> None:
+        if self._ftdi.is_connected:
             self._ftdi.close()
-            self._ftdi = None
 
-    def purge(self):
+    def purge(self) -> None:
         self._ftdi.purge_buffers()
 
-    def reset(self, sync=False):
+    def reset(self, sync: bool = False) -> None:
         """Reset the attached TAP controller.
            sync sends the command immediately (no caching)
         """
         # we can either send a TRST HW signal or perform 5 cycles with TMS=1
         # to move the remote TAP controller back to 'test_logic_reset' state
         # do both for now
-        if not self._ftdi:
+        if not self._ftdi.is_connected:
             raise JtagError("FTDI controller terminated")
         if self._trst:
             # nTRST
             value = 0
-            cmd = Array('B', [Ftdi.SET_BITS_LOW, value, self.direction])
+            cmd = bytearray((Ftdi.SET_BITS_LOW, value, self.direction))
             self._ftdi.write_data(cmd)
-            time.sleep(0.1)
+            sleep(0.1)
             # nTRST should be left to the high state
             value = JtagController.TRST_BIT
-            cmd = Array('B', [Ftdi.SET_BITS_LOW, value, self.direction])
+            cmd = bytearray((Ftdi.SET_BITS_LOW, value, self.direction))
             self._ftdi.write_data(cmd)
-            time.sleep(0.1)
+            sleep(0.1)
         # TAP reset (even with HW reset, could be removed though)
         self.write_tms(BitSequence('11111'))
         if sync:
             self.sync()
 
-    def sync(self):
-        if not self._ftdi:
+    def sync(self) -> None:
+        if not self._ftdi.is_connected:
             raise JtagError("FTDI controller terminated")
         if self._write_buff:
             self._ftdi.write_data(self._write_buff)
-            self._write_buff = Array('B')
+            self._write_buff = bytearray()
 
-    def write_tms(self, tms):
+    def write_tms(self, tms: BitSequence) -> None:
         """Change the TAP controller state"""
         if not isinstance(tms, BitSequence):
             raise JtagError('Expect a BitSequence')
         length = len(tms)
-        if not (0 < length < 8):
+        if not 0 < length < 8:
             raise JtagError('Invalid TMS length')
         out = BitSequence(tms, length=8)
         # apply the last TDO bit
         if self._last is not None:
             out[7] = self._last
-        # print "TMS", tms, (self._last is not None) and 'w/ Last' or ''
+        # print("TMS", tms, (self._last is not None) and 'w/ Last' or '')
         # reset last bit
         self._last = None
-        cmd = Array('B', [Ftdi.WRITE_BITS_TMS_NVE, length-1, out.tobyte()])
+        cmd = bytearray((Ftdi.WRITE_BITS_TMS_NVE, length-1, out.tobyte()))
         self._stack_cmd(cmd)
         self.sync()
-        #self._ftdi.validate_mpsse()
 
-    def read(self, length):
-        """Read out a sequence of bits from TDO"""
+    def read(self, length: int) -> BitSequence:
+        """Read out a sequence of bits from TDO."""
         byte_count = length//8
         bit_count = length-8*byte_count
         bs = BitSequence()
         if byte_count:
-            bytes = self._read_bytes(byte_count)
-            bs.append(bytes)
+            bytes_ = self._read_bytes(byte_count)
+            bs.append(bytes_)
         if bit_count:
             bits = self._read_bits(bit_count)
             bs.append(bits)
         return bs
 
-    def write(self, out, use_last=True):
+    def write(self, out: Union[BitSequence, str], use_last: bool = True):
         """Write a sequence of bits to TDI"""
         if isinstance(out, str):
+            if len(out) > 1:
+                self._write_bytes_raw(out[:-1])
+                out = out[-1]
             out = BitSequence(bytes_=out)
         elif not isinstance(out, BitSequence):
             out = BitSequence(out)
@@ -293,7 +302,8 @@ class JtagController(object):
         if bit_count:
             self._write_bits(out[pos:])
 
-    def shift_register(self, out, use_last=False):
+    def shift_register(self, out: BitSequence,
+                       use_last: bool = False) -> BitSequence:
         """Shift a BitSequence into the current register and retrieve the
            register output"""
         if not isinstance(out, BitSequence):
@@ -308,17 +318,18 @@ class JtagController(object):
             raise JtagError("Nothing to shift")
         if byte_count:
             blen = byte_count-1
-            #print "RW OUT %s" % out[:pos]
-            cmd = Array('B', [Ftdi.RW_BYTES_PVE_NVE_LSB, blen, (blen>>8)&0xff])
+            # print("RW OUT %s" % out[:pos])
+            cmd = bytearray((Ftdi.RW_BYTES_PVE_NVE_LSB,
+                             blen, (blen >> 8) & 0xff))
             cmd.extend(out[:pos].tobytes(msby=True))
             self._stack_cmd(cmd)
-            #print "push %d bytes" % byte_count
+            # print("push %d bytes" % byte_count)
         if bit_count:
-            #print "RW OUT %s" % out[pos:]
-            cmd = Array('B', [Ftdi.RW_BITS_PVE_NVE_LSB, bit_count-1])
+            # print("RW OUT %s" % out[pos:])
+            cmd = bytearray((Ftdi.RW_BITS_PVE_NVE_LSB, bit_count-1))
             cmd.append(out[pos:].tobyte())
             self._stack_cmd(cmd)
-            #print "push %d bits" % bit_count
+            # print("push %d bits" % bit_count)
         self.sync()
         bs = BitSequence()
         byte_count = length//8
@@ -329,9 +340,9 @@ class JtagController(object):
             if not data:
                 raise JtagError('Unable to read data from FTDI')
             byteseq = BitSequence(bytes_=data, length=8*byte_count)
-            #print "RW IN %s" % byteseq
+            # print("RW IN %s" % byteseq)
             bs.append(byteseq)
-            #print "pop %d bytes" % byte_count
+            # print("pop %d bytes" % byte_count)
         if bit_count:
             data = self._ftdi.read_data_bytes(1, 4)
             if not data:
@@ -341,15 +352,22 @@ class JtagController(object):
             byte >>= 8-bit_count
             bitseq = BitSequence(byte, length=bit_count)
             bs.append(bitseq)
-            #print "pop %d bits" % bit_count
+            # print("pop %d bits" % bit_count)
         if len(bs) != length:
-            raise AssertionError("Internal error")
-        #self._ftdi.validate_mpsse()
+            raise ValueError("Internal error")
         return bs
 
-    def _stack_cmd(self, cmd):
-        if not isinstance(cmd, Array):
-            raise TypeError('Expect a byte array')
+    @property
+    def ftdi(self) -> Ftdi:
+        """Return the Ftdi instance.
+
+           :return: the Ftdi instance
+        """
+        return self._ftdi
+
+    def _stack_cmd(self, cmd: Union[bytes, bytearray]):
+        if not isinstance(cmd, (bytes, bytearray)):
+            raise TypeError('Expect bytes or bytearray')
         if not self._ftdi:
             raise JtagError("FTDI controller terminated")
         # Currrent buffer + new command + send_immediate
@@ -357,84 +375,95 @@ class JtagController(object):
             self.sync()
         self._write_buff.extend(cmd)
 
-    def _read_bits(self, length):
+    def _read_bits(self, length: int):
         """Read out bits from TDO"""
-        data = ''
         if length > 8:
             raise JtagError("Cannot fit into FTDI fifo")
-        cmd = Array('B', [Ftdi.READ_BITS_NVE_LSB, length-1])
+        cmd = bytearray((Ftdi.READ_BITS_NVE_LSB, length-1))
         self._stack_cmd(cmd)
         self.sync()
         data = self._ftdi.read_data_bytes(1, 4)
         # need to shift bits as they are shifted in from the MSB in FTDI
-        byte = ord(data) >> 8-bit_count
+        byte = data[0] >> 8-length
         bs = BitSequence(byte, length=length)
-        #print "READ BITS %s" % (bs)
+        # print("READ BITS %s" % bs)
         return bs
 
-    def _write_bits(self, out):
+    def _write_bits(self, out: BitSequence) -> None:
         """Output bits on TDI"""
         length = len(out)
         byte = out.tobyte()
-        #print "WRITE BITS %s" % out
-        cmd = Array('B', [Ftdi.WRITE_BITS_NVE_LSB, length-1, byte])
+        # print("WRITE BITS %s" % out)
+        cmd = bytearray((Ftdi.WRITE_BITS_NVE_LSB, length-1, byte))
         self._stack_cmd(cmd)
 
-    def _read_bytes(self, length):
+    def _read_bytes(self, length: int) -> BitSequence:
         """Read out bytes from TDO"""
-        data = ''
         if length > JtagController.FTDI_PIPE_LEN:
             raise JtagError("Cannot fit into FTDI fifo")
         alen = length-1
-        cmd = Array('B', [Ftdi.READ_BYTES_NVE_LSB, alen&0xff, (alen>>8)&0xff])
+        cmd = bytearray((Ftdi.READ_BYTES_NVE_LSB, alen & 0xff,
+                         (alen >> 8) & 0xff))
         self._stack_cmd(cmd)
         self.sync()
         data = self._ftdi.read_data_bytes(length, 4)
         bs = BitSequence(bytes_=data, length=8*length)
-        #print "READ BYTES %s" % bs
+        # print("READ BYTES %s" % bs)
         return bs
 
-    def _write_bytes(self, out):
+    def _write_bytes(self, out: BitSequence):
         """Output bytes on TDI"""
-        bytes_ = out.tobytes(msby=True) # don't ask...
+        bytes_ = out.tobytes(msby=True)  # don't ask...
         olen = len(bytes_)-1
-        #print "WRITE BYTES %s" % out
-        cmd = Array('B', [Ftdi.WRITE_BYTES_NVE_LSB, olen&0xff, (olen>>8)&0xff])
+        # print("WRITE BYTES %s" % out)
+        cmd = bytearray((Ftdi.WRITE_BYTES_NVE_LSB, olen & 0xff,
+                          (olen >> 8) & 0xff))
         cmd.extend(bytes_)
         self._stack_cmd(cmd)
 
+    def _write_bytes_raw(self, out: BitSequence):
+        """Output bytes on TDI"""
+        olen = len(out)-1
+        cmd = bytearray((Ftdi.WRITE_BYTES_NVE_LSB, olen & 0xff,
+                         (olen >> 8) & 0xff))
+        cmd.extend(out)
+        self._stack_cmd(cmd)
 
-class JtagEngine(object):
+
+class JtagEngine:
     """High-level JTAG engine controller"""
 
-    def __init__(self, trst=False, frequency=3E06):
+    def __init__(self, trst: bool = False, frequency: float = 3E06):
         self._ctrl = JtagController(trst, frequency)
         self._sm = JtagStateMachine()
-        self._seq = ''
+        self._seq = bytearray()
 
-    def debug(self, level):
-        """Change the current debug level"""
-        self._ctrl.debug(level)
+    @property
+    def state_machine(self):
+        return self._sm
 
-    def configure(self, vendor=0x0403, product=0x6011, interface=0):
+    @property
+    def controller(self):
+        return self._ctrl
+
+    def configure(self, url: str) -> None:
         """Configure the FTDI interface as a JTAG controller"""
-        self._ctrl.configure(vendor=vendor, product=product,
-                             interface=interface)
+        self._ctrl.configure(url)
 
-    def close(self):
+    def close(self) -> None:
         """Terminate a JTAG session/connection"""
         self._ctrl.close()
 
-    def purge(self):
+    def purge(self) -> None:
         """Purge low level HW buffers"""
         self._ctrl.purge()
 
-    def reset(self):
+    def reset(self) -> None:
         """Reset the attached TAP controller"""
         self._ctrl.reset()
         self._sm.reset()
 
-    def write_tms(self, out):
+    def write_tms(self, out) -> None:
         """Change the TAP controller state"""
         self._ctrl.write_tms(out)
 
@@ -442,7 +471,7 @@ class JtagEngine(object):
         """Read out a sequence of bits from TDO"""
         return self._ctrl.read(length)
 
-    def write(self, out, use_last=False):
+    def write(self, out, use_last=False) -> None:
         """Write a sequence of bits to TDI"""
         self._ctrl.write(out, use_last)
 
@@ -450,7 +479,7 @@ class JtagEngine(object):
         """Return a list of supported state name"""
         return [str(s) for s in self._sm.states]
 
-    def change_state(self, statename):
+    def change_state(self, statename) -> None:
         """Advance the TAP controller to the defined state"""
         # find the state machine path to move to the new instruction
         path = self._sm.find_path(statename)
@@ -461,38 +490,38 @@ class JtagEngine(object):
         # update the current state machine's state
         self._sm.handle_events(events)
 
-    def go_idle(self):
+    def go_idle(self) -> None:
         """Change the current TAP controller to the IDLE state"""
         self.change_state('run_test_idle')
 
-    def write_ir(self, instruction):
+    def write_ir(self, instruction) -> None:
         """Change the current instruction of the TAP controller"""
         self.change_state('shift_ir')
         self._ctrl.write(instruction)
         self.change_state('update_ir')
 
-    def capture_ir(self):
+    def capture_ir(self) -> None:
         """Capture the current instruction from the TAP controller"""
         self.change_state('capture_ir')
 
-    def write_dr(self, data):
+    def write_dr(self, data) -> None:
         """Change the data register of the TAP controller"""
         self.change_state('shift_dr')
         self._ctrl.write(data)
         self.change_state('update_dr')
 
-    def read_dr(self, length):
+    def read_dr(self, length: int) -> BitSequence:
         """Read the data register from the TAP controller"""
         self.change_state('shift_dr')
         data = self._ctrl.read(length)
         self.change_state('update_dr')
         return data
 
-    def capture_dr(self):
+    def capture_dr(self) -> None:
         """Capture the current data register from the TAP controller"""
         self.change_state('capture_dr')
 
-    def shift_register(self, length):
+    def shift_register(self, length) -> BitSequence:
         if not self._sm.state_of('shift'):
             raise JtagError("Invalid state: %s" % self._sm.state())
         if self._sm.state_of('capture'):
@@ -501,22 +530,22 @@ class JtagEngine(object):
             self._sm.handle_events(bs)
         return self._ctrl.shift_register(length)
 
-    def sync(self):
+    def sync(self) -> None:
         self._ctrl.sync()
 
 
-class JtagTool(object):
+class JtagTool:
     """A helper class with facility functions"""
 
     def __init__(self, engine):
         self._engine = engine
 
-    def idcode(self):
-        idcode = self.read_dr(32)
+    def idcode(self) -> None:
+        idcode = self._engine.read_dr(32)
         self._engine.go_idle()
         return int(idcode)
 
-    def preload(self, bsdl, data):
+    def preload(self, bsdl, data) -> None:
         instruction = bsdl.get_jtag_ir('preload')
         self._engine.write_ir(instruction)
         self._engine.write_dr(data)
@@ -529,36 +558,36 @@ class JtagTool(object):
         self._engine.go_idle()
         return data
 
-    def extest(self, bsdl):
+    def extest(self, bsdl) -> None:
         instruction = bsdl.get_jtag_ir('extest')
         self._engine.write_ir(instruction)
 
     def readback(self, bsdl):
-        data = self.read_dr(bsdl.get_boundary_length())
+        data = self._engine.read_dr(bsdl.get_boundary_length())
         self._engine.go_idle()
         return data
 
-    def detect_register_size(self):
+    def detect_register_size(self) -> int:
         # Freely inpired from UrJTAG
-        # Need to contact authors, or to replace this code to comply with
-        # the GPL license (GPL vs. LGPL with Python is a bit fuzzy for me)
-        if not self._engine._sm.state_of('shift'):
-            raise JtagError("Invalid state: %s" % self._engine._sm.state())
-        if self._engine._sm.state_of('capture'):
+        stm = self._engine.state_machine
+        if not stm.state_of('shift'):
+            raise JtagError("Invalid state: %s" % stm.state())
+        if stm.state_of('capture'):
             bs = BitSequence(False)
-            self._engine._ctrl.write_tms(bs)
-            self._engine._sm.handle_events(bs)
+            self._engine.controller.write_tms(bs)
+            stm.handle_events(bs)
         MAX_REG_LEN = 1024
         PATTERN_LEN = 8
         stuck = None
         for length in range(1, MAX_REG_LEN):
-            print "Testing for length %d" % length
+            print("Testing for length %d" % length)
             if length > 5:
                 return
             zero = BitSequence(length=length)
             inj = BitSequence(length=length+PATTERN_LEN)
             inj.inc()
-            for p in range(1, 1<<PATTERN_LEN):
+            ok = False
+            for p in range(1, 1 << PATTERN_LEN):
                 ok = False
                 self._engine.write(zero, False)
                 rcv = self._engine.shift_register(inj)
@@ -577,7 +606,7 @@ class JtagTool(object):
                     break
                 inj.inc()
             if ok:
-                print "Register detected length: %d" % length
+                print("Register detected length: %d" % length)
                 return length
         if stuck is not None:
             raise JtagError('TDO seems to be stuck')
